using UnityEngine;
using System;
using UnityEngine.EventSystems;
using System.Collections.Generic;
using UnityEngine.UI;
using System.Linq;
using Leap.Unity;
using System.Text;
using Battlehub.RTSaveLoad;

namespace Ruvik
{
    class PHHands
    {
        static public JSONStorable atomPlugin;
        public static List<JointSelector> activeJoints = new List<JointSelector>();
        static List<string> usedAtoms = new List<string>();
        public static bool canPoseHands;
        public static Dictionary<string,List<string>> retrivedJoints= new Dictionary<string, List<string>>();
        public static bool warned = false;
        public static Dictionary<string,bool> Retrivedside= new Dictionary<string, bool>();
        private static StringBuilder outputBuilder = new StringBuilder();

        public class RetrivedJoint
        {
            public string atomName;
            public List<string> joints;
            public RetrivedJoint(string name,List<string> joints)
            {
                atomName = name;
                this.joints = joints;
            }
        }
        public static void UpdateRetrivedSides()
        {
            if (PHUtils.activeAtom == null) return;
            Retrivedside[PHUtils.activeAtom.name] = PHUI.isRightSide;
        }
        public static void LoadPose(string path)
        {
            if (path.Length == 0) return;
            if (path.EndsWith(".jpg"))
                path = path.Substring(0, path.Length - "jpg".Length) + "json";
            //read file
            string LoadedHandPose = SuperController.singleton.ReadFileIntoString(path);
            string[] LoadedHandPoseLines = LoadedHandPose.Split(new string[] { "\r\n", "\n" }, StringSplitOptions.None);
            if (PHUI.isRightSide)
                foreach (string s in LoadedHandPoseLines)
                {
                    if (s.StartsWith("#"))
                        continue;
                    if (s.Length == 0)
                        continue;
                    string[] segValues = s.Split(new string[] { ":" }, StringSplitOptions.None);
                    string[] segmentID = segValues[0].Split(new string[] { " " }, StringSplitOptions.None);
                    string[] values = segValues[1].Split(new string[] { " " }, StringSplitOptions.None);
                    if (segmentID[0] == "Middle")
                        segmentID[0] = "Mid";

                    JointSelector js = GameObject.Find($"PH_node_{segmentID[0]}{segmentID[1]}").GetComponent<JointSelector>();

                    js.joints[0].SetRotation(new Vector3(float.Parse(values[0]), float.Parse(values[1]), float.Parse(values[2])));
                }
            else
                foreach (string s in LoadedHandPoseLines)
                {
                    if (s.StartsWith("#"))
                        continue;
                    if (s.Length == 0)
                        continue;
                    string[] segValues = s.Split(new string[] { ":" }, StringSplitOptions.None);
                    string[] segmentID = segValues[0].Split(new string[] { " " }, StringSplitOptions.None);
                    string[] values = segValues[1].Split(new string[] { " " }, StringSplitOptions.None);
                    if (segmentID[0] == "Middle")
                        segmentID[0] = "Mid";

                    JointSelector js = GameObject.Find($"PH_node_{segmentID[0]}{segmentID[1]}").GetComponent<JointSelector>();

                    js.joints[1].SetRotation(new Vector3(float.Parse(values[0]), -float.Parse(values[1]), -float.Parse(values[2])));
                }

        }
        public static void SavePose(string path)
        {
            if (path.Length == 0) return;
            outputBuilder.Length = 0;
            if (PHUI.isRightSide)
                foreach (var js in GameObject.FindObjectsOfType<JointSelector>())
                {
                    if (js.joints[0].joint.baseJointRotation == Vector3.zero)
                        continue;
                    if (js.jointID == "Middle")
                        js.jointID = "Mid";
                    outputBuilder.AppendLine(js.jointID + " " + js.jointNumber + ":" + js.joints[0].joint.baseJointRotation.x + " " + js.joints[0].joint.baseJointRotation.y + " " + js.joints[0].joint.baseJointRotation.z);
                }
            else
                foreach (var js in GameObject.FindObjectsOfType<JointSelector>())
                {
                    if (js.joints[1].joint.baseJointRotation == Vector3.zero)
                        continue;
                    if (js.jointID == "Middle")
                        js.jointID = "Mid";
                    outputBuilder.AppendLine(js.jointID + " " + js.jointNumber + ":" + js.joints[1].joint.baseJointRotation.x + " " + -js.joints[1].joint.baseJointRotation.y + " " + -js.joints[1].joint.baseJointRotation.z);
                }
            if (!path.EndsWith(".json"))
            {
                path += ".json";
            }

            SaveFile(path);
        }
        protected static bool SaveFile(string path)
        {
            if (SuperController.singleton.GetDirectoriesAtPath(SuperController.singleton.savesDir).Contains(@"Saves\\HandPose"))
            {
                SuperController.singleton.DoSaveScreenshot(SuperController.singleton.savesDir + "HandPose\\" + path.Substring(0, path.Length - "json".Length) + "jpg");
                SuperController.singleton.SaveStringIntoFile(SuperController.singleton.savesDir + "HandPose\\" + path, outputBuilder.ToString());
            }else
            {
                SuperController.LogError(SuperController.singleton.savesDir + "HandPose\\" + " does not exist, to save presets you need to create the \"HandPose\" directory inside the \"Saves\" directory");
            }
            return true;
        }
        public static void UpdateSelectors()
        {
            if (!PHUtils.activeAtom.GetStorableIDs().Any((storable) => storable.Contains("PHFingerSliders")))
            {
                if (!warned)
                {
                    SuperController.LogMessage("to pose fingers you need to add the \"PHFingerSlider.cs\" to the atom");
                    warned = true;
                }
                if (!PHUI.isBodyPosing)
                    PHUI.ToggleBodyPosing();
                return;
            }
            atomPlugin = PHUtils.activeAtom.GetStorableByID(PHUtils.activeAtom.GetStorableIDs().Find((storable) => storable.Contains("PHFingerSliders")));

            ClearActiveJoints();
            PHUI.phFingerPosingMenu.SetActive(true);

            UpdateSelector("Thumb", 0);
            UpdateSelector("Carpal", 0);
            UpdateSelector("Index", 1);
            UpdateSelector("Mid", 1);
            UpdateSelector("Ring", 2);
            UpdateSelector("Pinky", 2);

            if (!usedAtoms.Contains(PHUtils.activeAtom.name))
                usedAtoms.Add(PHUtils.activeAtom.name);

            PHUtils.activeAtom.GetStorableByID("LeftHandControl").SetStringChooserParamValue("fingerControlMode", "Morphs");
            PHUtils.activeAtom.GetStorableByID("RightHandControl").SetStringChooserParamValue("fingerControlMode", "Morphs");

            if (Retrivedside.ContainsKey(PHUtils.activeAtom.name))
                if (Retrivedside.Get(PHUtils.activeAtom.name) != PHUI.isRightSide)
                {
                    PHUI.SwitchSides();
                }
            RetriveActiveJoints();

            PHUI.phFingerPosingMenu.SetActive(!PHUI.isBodyPosing);
        }
        public static void UpdateRetrivedJoints()
        {
            if (PHUtils.activeAtom == null) return;
            retrivedJoints[PHUtils.activeAtom.name + PHUI.isRightSide.ToString()] = activeJoints.Select(js => js.jointID + js.jointNumber).ToList();
        }
        static void RetriveActiveJoints()
        {
            if (retrivedJoints.ContainsKey(PHUtils.activeAtom.name + PHUI.isRightSide.ToString()))
            {
                foreach (var item in retrivedJoints[PHUtils.activeAtom.name + PHUI.isRightSide.ToString()])
                {
                    var js = GetJointSelectorByName(item);
                    if (!activeJoints.Contains(js))
                    {
                        activeJoints.Add(js);
                        js.tex.color = Color.green;
                    }

                }

            }
        }
        public static JointSelector GetJointSelectorByName(string name)
        {
            return GameObject.Find($"PH_node_{name}").GetComponent<JointSelector>();
        }
        private static void InitFingerSelector(string name)
        {
            int count = 3;
            if (name == "Carpal")
                count = 2;
            for (int i = 1; i <= count; i++)
            {
                GameObject.Find($"PH_node_{name}{i}").AddComponent<JointSelector>().SetName();
            }
        }
        public static void InitFingerSelectors()
        {
            InitFingerSelector("Thumb");
            InitFingerSelector("Carpal");
            InitFingerSelector("Index");
            InitFingerSelector("Mid");
            InitFingerSelector("Ring");
            InitFingerSelector("Pinky");
        }

        public static void SwitchSides()
        {
            ClearActiveJoints();
            PHUI.SwitchSides();
            RetriveActiveJoints();
        }
        public static void UpdateSelector(string name, int carpal)
        {
            int count = 3;
            if (name == "Carpal")
                count = 2;
            for (int i = 1; i <= count; i++)
            {
                GameObject.Find($"PH_node_{name}{i}").GetComponent<JointSelector>().UpdateName(carpal);
            }
        }

        public static void ClearActiveJoints()
        {
            foreach (var joint in activeJoints)
            {
                joint.Clear();
            }
            activeJoints.Clear();

        }
        public class FingerJoint
        {
            public DAZBone joint;
            public JSONStorableFloat[] sliders;
            public FingerJoint(char side, string fingerName, int carpal)
            {
                var gender = PHUtils.activeAtom.transform.Find("rescale2/PhysicsModel/Genesis2Male") != null ? "Male" : "Female";
                sliders = new JSONStorableFloat[3];
                if (fingerName.Contains("Carpal"))
                {
                    joint = PHUtils.activeAtom.transform.Find($"rescale2/PhysicsModel/Genesis2{gender}/hip/abdomen/abdomen2/chest/{side}Collar/{side}Shldr/{side}ForeArm/{side}Hand/{side}{fingerName}").GetComponent<DAZBone>();
                }
                else
                {

                    int index = (int)fingerName[fingerName.Length - 1] - 48;
                    string NewfingerName = fingerName.Substring(0, fingerName.Length - 1);
                    string name = $"hip/abdomen/abdomen2/chest/{side}Collar/{side}Shldr/{side}ForeArm/{side}Hand/";

                    if (carpal > 0) name += $"{side}Carpal{carpal}/";

                    for (int i = 1; i <= index; i++)
                    {
                        name += $"{side}{NewfingerName}{i}/";
                    }
                    joint = PHUtils.activeAtom.transform.Find($"rescale2/PhysicsModel/Genesis2{gender}/{name}").GetComponent<DAZBone>();
                }

                sliders[0] = atomPlugin.GetFloatJSONParam($"{side}{fingerName}X"); sliders[0].valNoCallback = joint.baseJointRotation.x;
                sliders[1] = atomPlugin.GetFloatJSONParam($"{side}{fingerName}Y"); sliders[1].valNoCallback = joint.baseJointRotation.y;
                sliders[2] = atomPlugin.GetFloatJSONParam($"{side}{fingerName}Z"); sliders[2].valNoCallback = joint.baseJointRotation.z;
                if (usedAtoms.Contains(PHUtils.activeAtom.name))
                    return;
                sliders[0].slider.onValueChanged.AddListener((f) => SetRotationX(f));
                sliders[1].slider.onValueChanged.AddListener((f) => SetRotationY(f));
                sliders[2].slider.onValueChanged.AddListener((f) => SetRotationZ(f));
            }
            public void SetRotation(Vector3 rot)
            {
                sliders[0].val = rot.x;
                sliders[1].val = rot.y;
                sliders[2].val = rot.z;
            }
            public void SetRotationX(float f)
            {
                joint.baseJointRotation = new Vector3(f, joint.baseJointRotation.y, joint.baseJointRotation.z);
            }
            public void SetRotationY(float f)
            {
                joint.baseJointRotation = new Vector3(joint.baseJointRotation.x, f, joint.baseJointRotation.z);

            }
            public void SetRotationZ(float f)
            {
                joint.baseJointRotation = new Vector3(joint.baseJointRotation.x, joint.baseJointRotation.y, f);

            }
            public void Rotate(Vector3 rot)
            {
                sliders[0].val = joint.baseJointRotation.x > 120.0f ? 120.0f : joint.baseJointRotation.x < -120.0f ? -120.0f : joint.baseJointRotation.x + rot.x;
                sliders[1].val = joint.baseJointRotation.y > 120.0f ? 120.0f : joint.baseJointRotation.y < -120.0f ? -120.0f : joint.baseJointRotation.y + rot.y;
                sliders[2].val = joint.baseJointRotation.z > 120.0f ? 120.0f : joint.baseJointRotation.z < -120.0f ? -120.0f : joint.baseJointRotation.z + rot.z;
            }
            public void Reset()
            {
                joint.baseJointRotation = Vector3.zero;
                sliders[0].val = 0.0f;
                sliders[1].val = 0.0f;
                sliders[2].val = 0.0f;
            }
        }
        public class JointSelector : MonoBehaviour, IPointerEnterHandler, IPointerExitHandler
        {
            public FingerJoint[] joints = new FingerJoint[2];

            public delegate void RotateJoint(Vector3 dir);
            public delegate void SwapJoint();
            public Image tex;
            string jointName;
            public string jointID;
            public int jointNumber;
            bool isMoused = false;

            public void SetName()
            {
                jointName = gameObject.name.Split('_')[2];
                jointID = jointName.Substring(0, jointName.Length - 1);
                jointNumber = jointName[jointName.Length - 1] - 48;

                tex = GetComponent<Image>();
            }
            public void OnPointerEnter(PointerEventData eventData)
            {
                isMoused = true;
            }
            public void OnPointerExit(PointerEventData eventData)
            {
                isMoused = false;
            }
            public void UpdateName(int carpal)
            {
                joints[0] = new FingerJoint('r', jointName, carpal);
                joints[1] = new FingerJoint('l', jointName, carpal);
            }
            public void Mirror()
            {

                Vector3 altRot = joints[PHUI.isRightSide ? 0 : 1].joint.baseJointRotation;
                joints[PHUI.isRightSide ? 1 : 0].SetRotation(new Vector3(altRot.x, -altRot.y, -altRot.z));
            }
            void Update()
            {
                if (PHUI.isSymmetryToggled)
                    Mirror();

                if (!isMoused)
                    return;

                if (Input.GetMouseButtonDown(PHUI.MOUSE_LEFT))
                {
                    Select();
                    UpdateRetrivedJoints();
                }

                if (Input.GetMouseButtonDown(PHUI.MOUSE_RIGHT))
                {
                    if (Input.GetKey(KeyCode.LeftControl))
                        foreach (var item in FindObjectsOfType<JointSelector>())
                            if (item.jointID == this.jointID && item.jointNumber > this.jointNumber)
                                item.joints[PHUI.isRightSide ? 0 : 1].Reset();
                    joints[PHUI.isRightSide ? 0 : 1].Reset();
                }

                if (Input.GetMouseButtonDown(PHUI.MOUSE_MIDDLE))
                    Mirror();

            }
            public void Select()
            {
                if (!Input.GetKey(KeyCode.LeftShift))
                {
                    ClearActiveJoints();
                }
                if (Input.GetKey(KeyCode.LeftAlt))
                    if (this.jointID == "Carpal")
                        foreach (var item in FindObjectsOfType<JointSelector>().Where((js) => js.jointID == "Carpal"))
                        {
                            if (!activeJoints.Contains(item))
                            {
                                activeJoints.Add(item);
                                item.tex.color = Color.green;
                            }
                        }
                    else
                        foreach (var item in FindObjectsOfType<JointSelector>())
                            if (item.jointNumber == this.jointNumber && item.jointID != "Carpal" && !activeJoints.Contains(item))
                            {
                                activeJoints.Add(item);
                                item.tex.color = Color.green;
                            }
                if (Input.GetKey(KeyCode.LeftControl))
                    foreach (var item in FindObjectsOfType<JointSelector>())
                        if (item.jointID == this.jointID && item.jointNumber > this.jointNumber && !activeJoints.Contains(item))
                        {
                            activeJoints.Add(item);
                            item.tex.color = Color.green;
                        }
                if (!activeJoints.Contains(this))
                {
                    activeJoints.Add(this);
                    tex.color = Color.green;
                }
            }
            public void Clear()
            {
                tex.color = Color.white;
            }
        }
        public static void RotateActiveHandBones(Vector3 dir)
        {
            dir *= PH.sensetivity.valNoCallback;
            foreach (var joint in activeJoints)
            {
                joint.joints[PHUI.isRightSide ? 0 : 1].Rotate(dir);
            }
        }
    }
}
