using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace Ruvik
{
    class PHUndo
    {
        class BonesState
        {
            public Vector3 position;
            public Vector3 rotation;
            public FreeControllerV3.PositionState pstate;
            public FreeControllerV3.RotationState rstate;
            public BonesState(Vector3 position , Vector3 rotation , FreeControllerV3.PositionState pstate,FreeControllerV3.RotationState rstate)
            {
                this.position = position;
                this.rotation = rotation;
                this.pstate = pstate;
                this.rstate = rstate;
            }
        }
        static List<BonesState[]> states = new List<BonesState[]>();
        static List<Node> nodes = new List<Node>();
        public static void InitAtom()
        {
            Clear();
            var nss = GameObject.FindObjectsOfType<NodeSelector>();
            foreach (var item in nss)
            {
                nodes.Add(item.node);
                if (item.node.altNode != null)
                    nodes.Add(item.node.altNode);
            }
        }
        public static void Clear()
        {
            states.Clear();
        }
        static private void RestoreBone(FreeControllerV3 node,BonesState state)
        {
            if(node.transform.position != state.position)
            {
                node.MoveControl(state.position);
                node.currentPositionState = state.pstate;
            }
            if(node.transform.eulerAngles != state.rotation)
            {
                node.RotateControl(state.rotation);
                node.currentRotationState = state.rstate;
            }
        }
        static private BonesState StoreBone(FreeControllerV3 bone)
        {
            return new BonesState(bone.transform.position,bone.transform.eulerAngles,bone.currentPositionState,bone.currentRotationState);
        }
        static public void RestoreState()
        {
            if (states.Count <= 0)
                return;
            BonesState[] state = states.Last();
            for (int i = 0; i < nodes.Count; i++)
            {
                RestoreBone(nodes[i].controller, state[i]);
            }
            states.RemoveAt(states.Count-1);
        }
        static public void StoreState()
        {
            BonesState[] state = new BonesState[nodes.Count];

            for (int i = 0; i < nodes.Count; i++)
            {
                state[i] = new BonesState(nodes[i].controller.transform.position, nodes[i].controller.transform.eulerAngles, nodes[i].controller.currentPositionState, nodes[i].controller.currentRotationState);
            }

            if (states.Count >= 100)
                states.RemoveAt(0);
            states.Add(state);
        }
    }
}
