﻿using MVR.FileManagementSecure;
using Oculus.Platform.Models;
using SimpleJSON;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.CompilerServices;
using UnityEngine;

namespace Ruvik
{
    class PHUtils
    {
        static public FreeControllerV3 head;
        static public FreeControllerV3 neck;
        static public FreeControllerV3[] shoulders;
        static public FreeControllerV3[] elbows;
        static public FreeControllerV3[] hands;
        static public FreeControllerV3 chest;
        static public FreeControllerV3 abdomen2;
        static public FreeControllerV3 pelvis;
        static public FreeControllerV3 hip;
        static public FreeControllerV3 abdomen1;
        static public FreeControllerV3[] thighs;
        static public FreeControllerV3[] knees;
        static public FreeControllerV3[] feet;
        static public FreeControllerV3[] toes;
        static public FreeControllerV3[] penis;
        public static List<Node> activeNodes = new List<Node>();
        public static Atom activeAtom = null;
        static FreeControllerV3.PositionState[] currentPStates;
        static FreeControllerV3.RotationState[] currentRStates;
        public static Transform[] motionTransform;
        public static int motionTransformIndex = 0;
        static Transform middlePoint = null;

        public static float xPosMultiplier = 1.0f;
        public static float yPosMultiplier = 1.0f;
        public static float zPosMultiplier = 1.0f;
        public static float xRotMultiplier = 1.0f;
        public static float yRotMultiplier = 1.0f;
        public static float zRotMultiplier = 1.0f;

        static public void ToggleSymmetry()
        {
            PHUI.isSymmetryToggled = !PHUI.isSymmetryToggled;
            PHUI.symmetryImage.color = PHUI.isSymmetryToggled ? Color.green:Color.white;

            PHUI.ActivateRightSide(!PHUI.isSymmetryToggled);
            if(!PHUI.isBodyPosing)
                PHUI.GetButton("hand_side").interactable = !PHUI.isSymmetryToggled;
        }
        static public void ToggleState()
        {
            activeAtom.ToggleOn();
            PHUI.onOffImage.color = activeAtom.on ? Color.green : Color.white;
        }
        static public void ToggleFreeze()
        {
            activeAtom.SetFreezePhysics(!activeAtom.isPhysicsFrozen);
            PHUI.freezeImage.color = activeAtom.isPhysicsFrozen ? Color.green : Color.white;
        }
        static public void ToggleCollision()
        {
            activeAtom.SetCollisionEnabled(!activeAtom.collisionEnabled);
            PHUI.collisionImage.color = activeAtom.collisionEnabled ? Color.green : Color.white;
        }

        static string posePath = "Custom/Atom/Person/Pose";
        static List<ShortCut> poseShortcuts = FileManagerSecure.GetShortCutsForDirectory(posePath, false, false, true, true);
        static List<ShortCut> handShortcuts = FileManagerSecure.GetShortCutsForDirectory(SuperController.singleton.savesDir + "HandPose");
        static public void SavePose(string name, bool saveBody)
        {
            if (saveBody)
            {
                var posePresets = activeAtom.presetManagerControls.First(mc => mc.name == "PosePresets");
                posePresets.SetStringParamValue("presetName", name);
                posePresets.CallAction("StorePresetWithScreenshot");
            }
            else
            {
                PHHands.SavePose(name);
            }
        }
        static public void LoadPose(bool loadBody)
        {
            if (loadBody)
            {
                PHUndo.StoreState();
                SuperController.singleton.GetMediaPathDialog(HandleLoadPreset, "vap", posePath, false, true, false, null, false, poseShortcuts);
            }
            else
            {
                SuperController.singleton.NormalizeMediaPath(SuperController.singleton.savesDir + "HandPose");
                SuperController.singleton.GetMediaPathDialog((string path) =>
                {
                    PHHands.LoadPose(path);
                }, PH.useOldPresets.valNoCallback==true?null:"json", SuperController.singleton.savesDir + "HandPose", true, true, true, null, false, handShortcuts);
            }
        }
        static public void ResetPose_method()
        {
            PHUndo.StoreState();
            if (PHUI.isBodyPosing)
            {
                for (int i = 1; i < activeAtom.freeControllers.Length; i++)
                {
                    activeAtom.freeControllers[i].Reset();
                }

                if (PH.changePose.valNoCallback)
                {
                    var morphControl = ((activeAtom.GetStorableByID("geometry")) as DAZCharacterSelector).morphsControlUI;

                    morphControl.GetMorphUids().ForEach((uid) =>
                    {
                        var morph = morphControl.GetMorphByUid(uid);
                        if (morph.isPoseControl)
                        {
                            morph.SetDefaultValue();
                        }
                    });
                }
            }
            else
            {
                foreach (var joint in GameObject.FindObjectsOfType<PHHands.JointSelector>())
                {
                    joint.joints[PHUI.isRightSide ? 0 : 1].Reset();
                }
            }
        }

        static public void ChangeAlignment()
        {
            motionTransformIndex++;
            motionTransformIndex = motionTransformIndex % 3;
        }
        public static void UpdateStates()
        {
            if (activeAtom == null)
                return;

            PHUI.collisionImage.color = activeAtom.collisionEnabled ? Color.green : Color.white;
            PHUI.freezeImage.color = activeAtom.isPhysicsFrozen ? Color.green : Color.white;
            PHUI.onOffImage.color = activeAtom.on ? Color.green : Color.white;

            if (activeNodes != null && activeNodes.Count > 0)
                PHUI.UpdateStates(activeNodes.Last());
        }
        public static void UpdateNodes(Atom atom)
        {
            if (activeNodes == null)
                activeNodes = new List<Node>();


            PHUI.ClearActiveNodes();
            activeAtom = atom;
            motionTransform[0] = activeAtom.freeControllers[0].transform;
            knees = atom.freeControllers.Where(fc => fc.name.Contains("KneeControl")).ToArray();
            hands = atom.freeControllers.Where(fc => fc.name.Contains("HandControl")).ToArray();
            feet = atom.freeControllers.Where(fc => fc.name.Contains("FootControl")).ToArray();
            toes = atom.freeControllers.Where(fc => fc.name.Contains("ToeControl")).ToArray();
            thighs = atom.freeControllers.Where(fc => fc.name.Contains("ThighControl")).ToArray();
            elbows = atom.freeControllers.Where(fc => fc.name.Contains("ElbowControl")).ToArray();
            shoulders = atom.freeControllers.Where(fc => fc.name.Contains("ArmControl")).ToArray();
            penis = atom.freeControllers.Where(fc => fc.name.Contains("penis")).ToArray();
            hip = atom.freeControllers.First(fc => fc.name.Contains("hip"));
            chest = atom.freeControllers.First(fc => fc.name.Contains("chest"));
            head = atom.freeControllers.First(fc => fc.name.Contains("head"));
            neck = atom.freeControllers.First(fc => fc.name.Contains("neck"));
            abdomen1 = atom.freeControllers.First(fc => fc.name == "abdomenControl");
            abdomen2 = atom.freeControllers.First(fc => fc.name == "abdomen2Control");
            pelvis = atom.freeControllers.First(fc => fc.name == "pelvisControl");

            PHUI.phBodyPosingMenu.SetActive(true);
            PHUI.UpdateNode("head", head);
            PHUI.UpdateNode("neck", neck);
            PHUI.UpdateNode("chest", chest);
            PHUI.UpdateNode("abdomen2", abdomen2);
            PHUI.UpdateNode("hip", hip);
            PHUI.UpdateNode("pelvis", pelvis);
            PHUI.UpdateNode("abdomen", abdomen1);
            PHUI.UpdateNode("penisBase", penis[0]);
            PHUI.UpdateNode("penisMid", penis[1]);
            PHUI.UpdateNode("penisTip", penis[2]);
            PHUI.UpdateNode("control", activeAtom.freeControllers[0]);
            PHUI.UpdateNodeAlt("rToe", toes[0], "lToe", toes[1]);
            PHUI.UpdateNodeAlt("rFoot", feet[0], "lFoot", feet[1]);
            PHUI.UpdateNodeAlt("rKnee", knees[0], "lKnee", knees[1]);
            PHUI.UpdateNodeAlt("rThigh", thighs[0], "lThigh", thighs[1]);
            PHUI.UpdateNodeAlt("rArm", shoulders[0], "lArm", shoulders[1]);
            PHUI.UpdateNodeAlt("rElbow", elbows[0], "lElbow", elbows[1]);
            PHUI.UpdateNodeAlt("rHand", hands[0], "lHand", hands[1]);

            PHUndo.InitAtom();
            PHUI.phBodyPosingMenu.SetActive(PHUI.isBodyPosing);

            PHHands.UpdateSelectors();

            switch (PH.SAVE_MODE)
            {
                case 0: //keep
                    if (PHUI.RetrivedControllers.ContainsKey("keep"))
                        foreach (var str in PHUI.RetrivedControllers["keep"])
                            PHUI.GetNodeByName(str.Split('C')[0]).Select(true);
                    break;
                case 1: //retrieve
                    if (PHUI.RetrivedControllers.ContainsKey(activeAtom.name))
                        foreach (var str in PHUI.RetrivedControllers[activeAtom.name])
                            PHUI.GetNodeByName(str.Split('C')[0]).Select(true);
                    break;
                default:
                    break;
            };
        }

        public static void ResetLink_method()
        {
            PHUndo.StoreState();
            foreach (FreeControllerV3 fc in activeAtom.freeControllers)
            {
                fc.SelectLinkToRigidbody(null);
            }
            head.currentPositionState = FreeControllerV3.PositionState.On;
            head.currentRotationState = FreeControllerV3.RotationState.On;

            foreach (var fc in activeAtom.freeControllers.Where(fc => fc.name.Contains("ShoulderControl")))
            {
                fc.currentPositionState = FreeControllerV3.PositionState.Off;
                fc.currentRotationState = FreeControllerV3.RotationState.Off;
            }

            neck.currentPositionState = FreeControllerV3.PositionState.Off;
            neck.currentRotationState = FreeControllerV3.RotationState.Off;

            shoulders[0].currentPositionState = FreeControllerV3.PositionState.Off;
            shoulders[1].currentPositionState = FreeControllerV3.PositionState.Off;
            shoulders[0].currentRotationState = FreeControllerV3.RotationState.Off;
            shoulders[1].currentRotationState = FreeControllerV3.RotationState.Off;

            elbows[0].currentPositionState = FreeControllerV3.PositionState.Off;
            elbows[1].currentPositionState = FreeControllerV3.PositionState.Off;
            elbows[0].currentRotationState = FreeControllerV3.RotationState.Off;
            elbows[1].currentRotationState = FreeControllerV3.RotationState.Off;

            hands[0].currentPositionState = FreeControllerV3.PositionState.On;
            hands[1].currentPositionState = FreeControllerV3.PositionState.On;
            hands[0].currentRotationState = FreeControllerV3.RotationState.On;
            hands[1].currentRotationState = FreeControllerV3.RotationState.On;

            chest.currentPositionState = FreeControllerV3.PositionState.On;
            chest.currentRotationState = FreeControllerV3.RotationState.On;

            abdomen2.currentPositionState = FreeControllerV3.PositionState.Off;
            abdomen2.currentRotationState = FreeControllerV3.RotationState.Off;

            hip.currentPositionState = FreeControllerV3.PositionState.On;
            hip.currentRotationState = FreeControllerV3.RotationState.On;

            pelvis.currentPositionState = FreeControllerV3.PositionState.Off;
            pelvis.currentRotationState = FreeControllerV3.RotationState.Off;

            abdomen1.currentPositionState = FreeControllerV3.PositionState.Off;
            abdomen1.currentRotationState = FreeControllerV3.RotationState.Off;

            thighs[0].currentPositionState = FreeControllerV3.PositionState.Off;
            thighs[1].currentPositionState = FreeControllerV3.PositionState.Off;
            thighs[0].currentRotationState = FreeControllerV3.RotationState.Off;
            thighs[1].currentRotationState = FreeControllerV3.RotationState.Off;

            knees[0].currentPositionState = FreeControllerV3.PositionState.Off;
            knees[1].currentPositionState = FreeControllerV3.PositionState.Off;
            knees[0].currentRotationState = FreeControllerV3.RotationState.Off;
            knees[1].currentRotationState = FreeControllerV3.RotationState.Off;

            feet[0].currentPositionState = FreeControllerV3.PositionState.On;
            feet[1].currentPositionState = FreeControllerV3.PositionState.On;
            feet[0].currentRotationState = FreeControllerV3.RotationState.On;
            feet[1].currentRotationState = FreeControllerV3.RotationState.On;

            toes[0].currentPositionState = FreeControllerV3.PositionState.Off;
            toes[1].currentPositionState = FreeControllerV3.PositionState.Off;
            toes[0].currentRotationState = FreeControllerV3.RotationState.Off;
            toes[1].currentRotationState = FreeControllerV3.RotationState.Off;

            penis[0].currentPositionState = FreeControllerV3.PositionState.Off;
            penis[1].currentPositionState = FreeControllerV3.PositionState.Off;
            penis[2].currentPositionState = FreeControllerV3.PositionState.Off;
            penis[0].currentRotationState = FreeControllerV3.RotationState.Off;
            penis[1].currentRotationState = FreeControllerV3.RotationState.Off;
            penis[2].currentRotationState = FreeControllerV3.RotationState.Off;

            if (activeNodes.Count > 0)
                PHUI.UpdateStates(activeNodes.Last());
        }
        public static void MirrorNode(Node node)
        {
            PHUndo.StoreState();
            node.controller.currentPositionState = FreeControllerV3.PositionState.On;
            node.controller.currentRotationState = FreeControllerV3.RotationState.On;
            node.controller.MoveControl(GetInversePosition(node.altNode.controller.transform));
            node.controller.RotateTo(GetInverseRotation(node.altNode.controller.transform));
            node.controller.currentPositionState = node.altNode.controller.currentPositionState;
            node.controller.currentRotationState = node.altNode.controller.currentRotationState;
        }
        public static void SwapNode(Node node)
        {
            PHUndo.StoreState();
            Vector3 tempPos = GetInversePosition(node.controller.transform);
            Vector3 tempRot = GetInverseRotation(node.controller.transform).eulerAngles;
            FreeControllerV3.PositionState tempPState = node.controller.currentPositionState;
            FreeControllerV3.RotationState temRPState = node.controller.currentRotationState;
            node.controller.MoveControl(GetInversePosition(node.altNode.controller.transform));
            node.controller.RotateControl(GetInversePosition(node.altNode.controller.transform));
            node.controller.currentPositionState = node.altNode.controller.currentPositionState;
            node.controller.currentRotationState = node.altNode.controller.currentRotationState;
            node.altNode.controller.MoveControl(tempPos);
            node.altNode.controller.RotateControl(tempRot);
            node.altNode.controller.currentPositionState = tempPState;
            node.altNode.controller.currentRotationState = temRPState;
        }
        public static void PreMoveControllers()
        {
            PHUndo.StoreState();
            currentPStates = new FreeControllerV3.PositionState[activeNodes.Count];
            currentRStates = new FreeControllerV3.RotationState[activeNodes.Count];
            for (int i = 0; i < activeNodes.Count; i++)
            {
                currentPStates[i] = activeNodes[i].controller.currentPositionState;
                currentRStates[i] = activeNodes[i].controller.currentRotationState;
            }

            var activeNodesCopy = new List<Node>(activeNodes);
            activeNodesCopy.Sort(delegate (Node a, Node b)
            {
                if (a.controller.transform.position.y == b.controller.transform.position.y) return 0;
                if (a.controller.transform.position.y > b.controller.transform.position.y) return -1;
                if (a.controller.transform.position.y < b.controller.transform.position.y) return 1;
                else return a.controller.transform.position.y.CompareTo(b.controller.transform.position.y);
            });
            middlePoint = activeNodesCopy[activeNodesCopy.Count / 2].controller.transform;
        }
        public static void MoveActiveNodes(Vector3 dir, bool isChaingingPosition)
        {
            dir *= Time.deltaTime * PH.sensetivity.valNoCallback;
            if (isChaingingPosition)
            {
                foreach (var node in activeNodes)
                {
                    Transform motion = Input.GetKey(KeyCode.LeftShift) ? motionTransform[1] : Input.GetKey(KeyCode.LeftControl) ? motionTransform[2] : motionTransform[motionTransformIndex];
                    node.controller.currentPositionState = FreeControllerV3.PositionState.On;
                    node.controller.MoveControlRelatve(motion.right * (dir.x * xPosMultiplier) + motion.up * (dir.y * yPosMultiplier) + motion.forward * (dir.z * zPosMultiplier));
                    if (PHUI.isSymmetryToggled && node.altNode != null)
                    {
                        node.altNode.controller.currentPositionState = FreeControllerV3.PositionState.On;
                        node.altNode.controller.MoveControlRelatve(motion.right * (dir.x * -xPosMultiplier) + motion.up * (dir.y * yPosMultiplier) + motion.forward * (dir.z * zPosMultiplier));
                    }
                }
            }
            else
            {
                if (activeNodes.Count > 2 && Input.GetKey(KeyCode.LeftAlt))
                {
                    foreach (var node in activeNodes)
                    {
                        node.controller.currentRotationState = FreeControllerV3.RotationState.On;
                        node.controller.transform.RotateAround(middlePoint.transform.position, dir, Time.deltaTime * PH.rotationSpeed.valNoCallback * 20.0f);
                        if (PHUI.isSymmetryToggled && node.altNode != null)
                        {
                            node.altNode.controller.currentRotationState = FreeControllerV3.RotationState.On;
                            node.altNode.controller.transform.RotateAround(middlePoint.transform.position, dir, Time.deltaTime * PH.rotationSpeed.valNoCallback * 20.0f);
                        }
                    }
                }
                else
                {
                    foreach (var node in activeNodes)
                    {
                        Transform motion = Input.GetKey(KeyCode.LeftShift) ? motionTransform[1] : Input.GetKey(KeyCode.LeftControl) ? motionTransform[2] : node.controller.transform;
                        node.controller.currentRotationState = FreeControllerV3.RotationState.On;
                        node.controller.RotateControl(motion.right * (dir.x * xRotMultiplier) + motion.up * (dir.y * yRotMultiplier) + motion.forward * (dir.z * zRotMultiplier), PH.rotationSpeed.val);
                        if (PHUI.isSymmetryToggled && node.altNode != null)
                        {
                            Transform altMotion = Input.GetKey(KeyCode.LeftShift) ? motionTransform[1] : Input.GetKey(KeyCode.LeftControl) ? motionTransform[2] : node.altNode.controller.transform;
                            node.altNode.controller.currentRotationState = FreeControllerV3.RotationState.On;
                            node.altNode.controller.RotateControl(altMotion.right * (dir.x * xRotMultiplier) + altMotion.up * (-dir.y * yRotMultiplier) + altMotion.forward * (-dir.z * zRotMultiplier), PH.rotationSpeed.val);
                        }
                    }
                }
            }

        }
        public static void PostMoveControllers()
        {
            for (int i = 0; i < activeNodes.Count; i++)
            {
                activeNodes[i].controller.currentPositionState = currentPStates[i];
                activeNodes[i].controller.currentRotationState = currentRStates[i];
                if (PHUI.isSymmetryToggled && activeNodes[i].altNode != null)
                {
                    activeNodes[i].altNode.controller.currentPositionState = currentPStates[i];
                    activeNodes[i].altNode.controller.currentRotationState = currentRStates[i];
                }
            }
        }
        static Vector3 GetInversePosition(Transform trans)
        {
            Vector3 inversePos = Vector3.Reflect(trans.position - activeAtom.freeControllers[0].transform.position, activeAtom.freeControllers[0].transform.right);
            return activeAtom.freeControllers[0].transform.position + inversePos;
        }
        static Quaternion GetInverseRotation(Transform trans)
        {
            return ReflectRotation(trans.rotation, activeAtom.freeControllers[0].transform.right);
        }
        static Quaternion ReflectRotation(Quaternion source, Vector3 normal)
        {
            return Quaternion.LookRotation(Vector3.Reflect(source * Vector3.forward, normal), Vector3.Reflect(source * Vector3.up, normal));
        }
        public static void ClearLink()
        {
            foreach (var node in activeNodes)
            {
                node.controller.SelectLinkToRigidbody(null);
                node.controller.currentPositionState = FreeControllerV3.PositionState.On;
                node.controller.currentRotationState = FreeControllerV3.RotationState.On;
            }
            PHUI.UpdateStates(activeNodes.Last());
        }
        public static void IKLink()
        {
            switch (activeNodes.Count)
            {
                case 0:
                    IKLink(pelvis);
                    break;
                case 1:
                    if (activeNodes[0].controller.name == activeAtom.freeControllers[0].name)
                        IKLink(pelvis);
                    else
                        IKLink(activeNodes[0].controller);
                    break;
                case 2:
                    if (Input.GetKey(KeyCode.LeftAlt))
                        Link(activeNodes[0].controller, activeNodes[1].controller);
                    else
                        IKLink(activeNodes[0].controller, activeNodes[1].controller);
                    break;
                case 3:
                    Link(activeNodes[0].controller, activeNodes[2].controller);
                    Link(activeNodes[1].controller, activeNodes[2].controller);
                    break;
                default:
                    return;
            }
        }
        static void IKLink(FreeControllerV3 a, FreeControllerV3 b)
        {
            PHUndo.StoreState();
            PHLinker.LinkPath(PHLinker.FindPath(a, b));
        }
        static void IKLink(FreeControllerV3 center)
        {
            PHLinker.LinkPath(PHLinker.FindPath(toes[0], center));
            PHLinker.LinkPath(PHLinker.FindPath(toes[1], center));
            PHLinker.LinkPath(PHLinker.FindPath(head, center));
            PHLinker.LinkPath(PHLinker.FindPath(penis[2], center));
            PHLinker.LinkPath(PHLinker.FindPath(hands[1], center));
            PHLinker.LinkPath(PHLinker.FindPath(hands[0], center));
            SetDefaultLinkState(center);
        }
        static void SetDefaultLinkState(FreeControllerV3 center)
        {
            if (Input.GetKey(KeyCode.LeftShift))
            {
                head.currentPositionState = FreeControllerV3.PositionState.Off;
                penis[1].currentPositionState = FreeControllerV3.PositionState.Off;
                penis[2].currentPositionState = FreeControllerV3.PositionState.Off;
                for (int i = 0; i < 2; i++)
                {
                    shoulders[i].currentPositionState = FreeControllerV3.PositionState.Off;
                    elbows[i].currentPositionState = FreeControllerV3.PositionState.Off;
                    hands[i].currentPositionState = FreeControllerV3.PositionState.Off;
                    thighs[i].currentPositionState = FreeControllerV3.PositionState.Off;
                    knees[i].currentPositionState = FreeControllerV3.PositionState.Off;
                    feet[i].currentPositionState = FreeControllerV3.PositionState.Off;
                    toes[i].currentPositionState = FreeControllerV3.PositionState.Off;
                }
            }
            center.currentPositionState = FreeControllerV3.PositionState.On;
            center.currentRotationState = FreeControllerV3.RotationState.On;
        }
        public static void Link(FreeControllerV3 a, FreeControllerV3 b)
        {
            a.SelectLinkToRigidbody(b.gameObject.GetComponent<Rigidbody>());
            a.currentPositionState = FreeControllerV3.PositionState.ParentLink;
            a.currentRotationState = FreeControllerV3.RotationState.ParentLink;
        }
        static void HandleLoadPreset(string path)
        {
            var posePresets = activeAtom.presetManagerControls.First(mc => mc.name == "PosePresets");
            posePresets.CallPresetFileAction("LoadPresetWithPath", path);
        }
    }
}
